"""
Aibo client code to interface with Tekkotsu using only Python builtins
Adapted from Pyro.

Most code (c) 2005, Pyro Project. Licenced under the GNU GPL.
"""

__author__ = "Steven Kryskalla <skryskalla@gmail.com>"
__version__ = "0.6"

from listener import Listener, ListenerThread
from aiboftp import AiboFTP
from defs import dtor, clip, PORTS, JOINTS, BUTTONS, SENSORS, RANGES, POSNAMES

import struct, time

def makeControlCommand(control, amt):
    # HEAD: control is tilt 't', pan 'p', or roll 'r'
    # WALK: control is forward 'f', strafe 's', and rotate 'r'
    return struct.pack('<bf', ord(control), amt) 

##class AiboHeadDevice:
##    """Class for moving Aibo's head using pan, tilt, and roll
##    see HeadPointListener.java
##    TODO: get rid of this and put it in the AiboRobot class (see 'TODO: invert this')
##    to the user accessing the head and accessing the legs/tail/mouth should look the same"""
##    def __init__(self, robot):
##        self.robot = robot
##        self.robot.setRemoteControl("Head Remote Control", "on") #turn on head remote control
##        time.sleep(1) # pause a second for connection
##        self.dev = Listener(PORTS["Head Remote Control"], self.robot.host)
##
##    def setPose(self, *args):
##        if len(args) == 3:
##            pan, tilt, roll = args[0], args[1], args[2]
##        else:
##            raise AttributeError, "setPose takes pan, tilt, and roll"
##        self.pan(pan)
##        self.tilt(tilt)
##        self.roll(roll)
##
##    def tilt(self, amt):
##        self.dev.write(makeControlCommand('t', amt))
##
##    def pan(self, amt):
##        self.dev.write(makeControlCommand('p', amt)) 
##
##    def roll(self, amt):
##        self.dev.write(makeControlCommand('r', amt))

class AiboRobot:
    """Class representing the Aibo"""
    def __init__(self, host):
        self.host = host

        self.main_control = Listener(PORTS["Main Control"], self.host)
        self.main_control.s.send("!reset\n") # reset menu

        #throttle the send rate from Aibo to us (we can't handle much faster than 10 bits/second)
        #lower this if you can
        self.setAiboConfig("vision.rawcam_interval", 50)
        #self.setAiboConfig("vision.rle_interval", 50)
        #self.setAiboConfig("vision.worldState_interval", 50)

        self.setRemoteControl("Walk Remote Control", "on")
        self.setRemoteControl("EStop Remote Control", "on")
        self.setRemoteControl("World State Serializer", "on")
        self.setRemoteControl("Aibo 3D", "on")
        time.sleep(1) #pause for connection

        self.walk_control = Listener(PORTS["Walk Remote Control"], self.host)
        self.estop_control = Listener(PORTS["EStop Remote Control"], self.host) # emergency stop control
        self.sensor_socket = Listener(PORTS["World State Serializer"], self.host) # sensors
        self.sensor_thread = ListenerThread(self.sensor_socket, self.readWorldState)
        self.sensor_thread.start()        
        self.joint_socket = Listener(PORTS["Joint Writer"], self.host) # joints
        time.sleep(1) #pause for connection

        #connect to Aibo FTP server
        self.ftp = AiboFTP(host)
        
        self.estop_control.s.send("start\n")
        time.sleep(1) #pause for connection
        
        #start head controller device
        #TODO: invert this
        #self.ptz = AiboHeadDevice(self)

    def setAiboConfig(self, item, state):
        self.main_control.s.send("!set %s=%s\n" % (item, state))

    def setRemoteControl(self, item, state):
        # could also use "!root 'TekkotsuMon' %(item)", but that toggles on/off
        if state == "off":
            item = "#" + item
        self.main_control.s.send("!select \"%s\"\n" % item)

    def readWorldState(self, socket):
        """Used as a callback in ListenerThread for sockets that produce data too fast for us to read."""
        # read sensor/pid states:
        self.ws_timestamp = socket.read(4, "l")

        numPIDJoints = socket.read(4, "l")
        self.numPIDJoints = numPIDJoints
        self.positionRaw = socket.read(numPIDJoints * 4, "<%df" % numPIDJoints,all=1)

        numSensors = socket.read(4, "l")
        self.numSensors = numSensors
        self.sensorRaw = socket.read(numSensors * 4, "<%df" % numSensors,all=1)

        numButtons = socket.read(4, "l")
        self.numButtons = numButtons
        self.buttonRaw = socket.read(numButtons * 4, "<%df" % numButtons,all=1)
        
        self.dutyCycleRaw = socket.read(numPIDJoints * 4, "<%df" % numPIDJoints,all=1)
                
    def getJoint(self, query):
        """Get (position, dutyCycle) of joint by name"""
        if query in JOINTS:
            pos = JOINTS[query]
            return self.positionRaw[pos], self.dutyCycleRaw[pos]
        else:
            raise ValueError, "incorrect joint name: %s" % query

    def getButton(self, query):
        """Get value of button by name"""
        if query in BUTTONS:
            pos = BUTTONS[query]
            if query.startswith("paw") or query in ["chin", "wireless"]:
                #discrete buttons (0 or 1)
                return int(self.buttonRaw[pos])
            else:
                return self.buttonRaw[pos]
        else:
            raise ValueError, "incorrect button name: %s" % query

    def getSensor(self, query):
        """Get a sensor value"""
        if query in SENSORS:
            pos = SENSORS[query]
            return self.sensorRaw[pos]
        else:
            raise ValueError, "incorrect sensor name: %s" % query

    def setJoint(self, joint, value):
        """Set the position of a joint (value in radians)"""
        l = list(self.positionRaw)

        if joint in JOINTS:
            pos = JOINTS[joint]
            l[pos] = value
            self.joint_socket.write(struct.pack("<18f",*l))
            time.sleep(0.5)
        else:
            raise ValueError, "incorrect joint name: %s" % joint

    def connect(self):
        #TODO: this really is not a proper connect function
        self.estop_control.s.send("start\n")

    def disconnect(self):
        #TODO: and this really is not a proper disconnect function
        self.estop_control.s.send("stop\n")
        raise NotImplementedError("this does not actually disconnect")
       
    def setTargetVelocity(self, x, y, z):
        """Set the direction for the walk x (forward-back), y (left-right), z (rotate)"""
        #TODO: this is not yet a clone of the Tekkotsu behavior
        # x: -1 to 1 (backward to forward)
        # y: -1 to 1 (right to left)        
        # z: -1 to 1 (right to left)
        self.walk_control.write(makeControlCommand('f', x)) 
        self.walk_control.write(makeControlCommand('s', y))
        self.walk_control.write(makeControlCommand('r', z))

    def playSound(self, file):
        """Plays a sound file on the memory stick (.wav optional)"""
        file = file.upper();
        if (not file.endswith(".WAV")):
            file +=".WAV"
        self.main_control.s.send("!select \"%s\"\n" % "Play Sound")
        self.main_control.s.send("!select \"%s\"\n" % file) #select file

    #TODO: streamSound
        
    def setWalk(self, file):
        """Sets the walk parameters file (.prm optional), restarts walk controller"""
        file = file.upper();
        if (not file.endswith(".PRM")):
            file +=".PRM"
        self.walk_control.s.close()
        self.main_control.s.send("!select \"%s\"\n" % "Load Walk") # forces files to be read
        self.main_control.s.send("!select \"%s\"\n" % file) # select file
        self.main_control.s.send("!select \"%s\"\n" % "#WalkControllerBehavior") # turn off
        self.main_control.s.send("!select \"%s\"\n" % "-WalkControllerBehavior") # turn on
        # If you change the walk, then you have to reconnect
        # back onto the walk server
        time.sleep(2)
        self.walk_control = Listener(PORTS["Walk Remote Control"], self.host) # walk command

    def streamPos(self, filename):
        """Transfers a .pos file to the dog then immediately plays it"""
        self.ftp.transferPos(filename)
        filename = filename.split("\\")[-1]
        filename = filename.upper()
        time.sleep(0.5)
        self.main_control.s.send("!select \"%s\"\n" % "File Access")
        self.main_control.s.send("!select \"%s\"\n" % "Load Posture")
        self.main_control.s.send("!select \"%s\"\n" % filename)
        self.main_control.s.send("!reset\n")

    def streamMot(self, filename):
        """Transfers a .mot file to the dog then immediately plays it"""
        self.ftp.transferMot(filename)
        filename = filename.split("\\")[-1]
        filename = filename.upper()
        time.sleep(0.5)
        self.main_control.s.send("!select \"%s\"\n" % "File Access")
        self.main_control.s.send("!select \"%s\"\n" % "Run Motion Sequence")
        self.main_control.s.send("!select \"%s\"\n" % filename)
        self.main_control.s.send("!reset\n")

    def playMot(self, filename):
        """Plays a motion file from the /MS/DATA/MOTION directory"""
        filename = filename.upper()
        if not filename.endswith(".MOT"):
            filename = filename + ".MOT"
        self.main_control.s.send("!select \"%s\"\n" % "File Access")
        self.main_control.s.send("!select \"%s\"\n" % "Run Motion Sequence")
        self.main_control.s.send("!select \"%s\"\n" % filename)
        self.main_control.s.send("!reset\n")

    def playPos(self, filename):
        """Plays a posture file from the /MS/DATA/MOTION directory"""
        self.main_control.s.send("!select \"%s\"\n" % "File Access")
        self.main_control.s.send("!select \"%s\"\n" % "Load Posture")
        self.main_control.s.send("!select \"%s\"\n" % filename.upper())
        self.main_control.s.send("!reset\n")

    def reboot(self):
        """Issue reboot command to the dog"""
        self.main_control.s.send("!select \"%s\"\n" % "Shutdown?")
        self.main_control.s.send("!select \"%s\"\n" % "Reboot")

    def shutdown(self):
        """Issue shutdown command to the dog"""
        self.main_control.s.send("!select \"%s\"\n" % "Shutdown?")
        self.main_control.s.send("!select \"%s\"\n" % "Shutdown")

## TODO: relax is based on a behavior, either make it non-behavior, or make a PyTekkotsu set of binaries for Aibo (with standard functions such as relax)
##    def relax(self):
##        """Relax joints by setting all PID parameters to zero"""
##        self.main_control.s.send("!select \"%s\"\n" % "Mode Switch")
##        self.main_control.s.send("!select \"%s\"\n" % "RelaxBehavior")
##        self.main_control.s.send("!reset\n")
##
##    def unrelax(self):
##        """Unrelax joints"""
##        self.main_control.s.send("!select \"%s\"\n" % "Mode Switch")
##        self.main_control.s.send("!select \"%s\"\n" % "#RelaxBehavior")
##        self.main_control.s.send("!reset\n")

    def runModeSwitch(self, name):
        """Runs a behavior from mode switch"""
        self.main_control.s.send("!select \"%s\"\n" % "Mode Switch")
        self.main_control.s.send("!select \"%s\"\n" % name)

    def turnOffModeSwitch(self, name):
        """Turns off a behavior from mode switch"""
        name = "#" + name #behaviors that are running are prefixed with #
        self.main_control.s.send("!select \"%s\"\n" % "Mode Switch")
        self.main_control.s.send("!select \"%s\"\n" % name)


    def eStopOn(self):
        """Turns the E-Stop on."""
        self.estop_control.s.send("start\n")

    def eStopOff(self):
        """Turns the E-Stop off."""
        self.estop_control.s.send("stop\n")

    def capturePos(self, outfile):
        self.main_control.s.send("!select \"%s\"\n" % "File Access")
        self.main_control.s.send("!select \"%s\"\n" % "Save Posture")
        time.sleep(1.0)
        self.main_control.s.send("CAP.TMP\n")
        time.sleep(0.25)
        self.ftp.getText("/MS/DATA/MOTION/CAP.TMP", file(outfile, "w"))
        self.main_control.s.send("!reset\n")